/*  msl.h
 *
 *  Copyright (C) 2005 James Bergstra
 *
 *  This file is part of the Montreal Scientific Library.
 *
 *  The Montreal Scientific Library is free software; you can redistribute it
 *  and/or modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of the License,
 *  or (at your option) any later version.
 *
 *  The Montreal Scientific Library is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along with
 *  the Montreal Scientific Library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */


#ifndef LIBMSL_H
#define LIBMSL_H

#include <math.h>

#include <gsl/gsl_complex.h>
#include <gsl/gsl_complex_math.h>

#include <gsl/gsl_vector.h>
#include <gsl/gsl_vector_ulong.h>
#include <gsl/gsl_vector_complex_double.h>

#include <gsl/gsl_matrix.h>
#include <gsl/gsl_matrix_complex_double.h>

#include <gsl/gsl_rng.h>
#include <gsl/gsl_multimin.h>

#include <gsl/gsl_cblas.h>

#undef __BEGIN_DECLS
#undef __END_DECLS
#ifdef __cplusplus
#define __BEGIN_DECLS extern "C" {
#define __END_DECLS }
#else
#define __BEGIN_DECLS           /* empty */
#define __END_DECLS             /* empty */
#endif

__BEGIN_DECLS

////////////////////////////////////////////////
//
//Inline double-precision level-1 cblas_*
//

#define cblas_dswap msl_cblas_dswap
#define cblas_dcopy msl_cblas_dcopy
#define cblas_daxpy msl_cblas_daxpy
#define cblas_dscal msl_cblas_dscal
#define cblas_ddot msl_cblas_ddot
#define cblas_dnrm2 msl_cblas_dnrm2
#define cblas_dasum msl_cblas_dasum
#define cblas_idamax msl_cblas_idamax

extern inline void 
msl_cblas_dswap(const int N, double *X, const int incX, double *Y, const int incY)
{
    int i;
    double t;
    for (i = 0; i < N; ++i)
    {
        t = X[i*incX];
        X[i*incX] = Y[i*incY];
        Y[i*incY] = t;
    }
}
extern inline void 
msl_cblas_dcopy(const int N, const double *X, const int incX, double *Y, const int incY)
{
    int i;
    for (i = 0; i < N; ++i)
    {
        Y[i*incY] = X[i*incX];
    }
}
extern inline void
msl_cblas_daxpy(const int N, const double alpha, const double *X, const int incX, double *Y, const int incY)
{
    int i;
    for (i = 0; i < N; ++i)
    {
        Y[i*incY] += alpha * X[i*incX];
    }
}
extern inline void 
msl_cblas_dscal(const int N, const double alpha, double *X, const int incX)
{
    int i;
    for (i = 0; i < N; ++i)
    {
        X[i*incX] *= alpha;
    }
}
extern inline double 
msl_cblas_ddot(const int N, const double *X, const int incX, const double *Y, const int incY)
{
    int i;
    double rval = 0;
    for (i = 0; i < N; ++i)
    {
        rval += X[i*incX] * Y[i*incY];
    }
    return rval;
}
extern inline double
msl_cblas_dnrm2(const int N, const double *X, const int incX)
{
    int i;
    double rval = 0;
    for (i = 0; i < N; ++i)
    {
        rval += X[i*incX] * X[i*incX];
    }
    return sqrt(rval);
}
extern inline double
msl_cblas_dasum(const int N, const double *X, const int incX)
{
    int i;
    double rval = 0;
    for (i = 0; i < N; ++i)
    {
        rval += X[i*incX];
    }
    return rval;
}
extern inline CBLAS_INDEX 
msl_cblas_idamax(const int N, const double *X, const int incX)
{
    CBLAS_INDEX rval = 0;
    double m = *X;
    int i;
    for (i = 1; i < N; ++i)
    {
        if (X[i*incX] > m) {
            m = X[i*incX];
            rval = i;
        }
    }
    return rval;
}

///////////////////////////////////////////////
//
// selected 128-bit aligned gsl_*_alloc calls (for SIMD)
//

#define gsl_block_alloc msl_block_alloc
#define gsl_block_complex_alloc msl_block_complex_alloc

#define gsl_matrix_alloc msl_matrix_alloc
#define gsl_matrix_complex_alloc msl_matrix_complex_alloc

#define gsl_vector_alloc msl_vector_alloc
#define gsl_vector_complex_alloc msl_vector_complex_alloc

gsl_block * 
msl_block_alloc(size_t size);
gsl_block_complex * 
msl_block_complex_alloc(size_t size);

gsl_matrix *
msl_matrix_alloc(size_t size1, size_t size2);
gsl_matrix_complex *
msl_matrix_complex_alloc(size_t size1, size_t size2);

gsl_vector *
msl_vector_alloc(size_t size);
gsl_vector_complex *
msl_vector_complex_alloc(size_t size);

gsl_block * 
msl_block_realloc( gsl_block * b, size_t size );
#if 0
gsl_matrix * 
msl_matrix_realloc( gsl_matrix * m, size_t size1, size_t size2);
gsl_vector *
msl_vector_realloc( gsl_vector * m, size_t size);
#endif


//////////////////////////////////////////////////
//
// Additional operations on
// gsl_vector[_complex]
// gsl_matrix[_complex]
//

#define MSL_MATRIX_AT(m,i,j) ((m)->data[ (i) * (m)->tda + (j) ])
#define MSL_MATRIX_COMPLEX_AT(m,i,j) (* ((gsl_complex *)( (m)->data + (i) * (m)->tda * 2 + (j) * 2 )))
#define MSL_MATRIX_COMPLEX_AT_R(m,i,j) ((m)->data[ (i) * (m)->tda * 2 + (j) * 2 ])
#define MSL_MATRIX_COMPLEX_AT_I(m,i,j) ((m)->data[ (i) * (m)->tda * 2 + (j) * 2 + 1 ])

#define MSL_VECTOR_AT(v,i) ((v)->data[ (i) * (v)->stride ])
#define MSL_VECTOR_COMPLEX_AT(v,i) (* ((gsl_complex *)( (v)->data + (i) * (v)->stride * 2)))
#define MSL_VECTOR_COMPLEX_AT_R(v,i) ((v)->data[ (i) * (v)->stride * 2 ])
#define MSL_VECTOR_COMPLEX_AT_I(v,i) ((v)->data[ (i) * (v)->stride * 2 + 1 ])

#define MSL_MATRIX_SIZE_MATCH(a,b) (((a)->size1 == (b)->size1) && ((a)->size2 == (b)->size2))

int
msl_vector_check_isnan(const gsl_vector *);
int
msl_vector_check_isinf( const gsl_vector * );

void
msl_vector_print_line(char **, size_t *, const gsl_vector *);
void
msl_matrix_print_spreadsheet(char **, size_t *, const gsl_matrix * );

void  //set each element randomly, ~U(0,1)
msl_matrix_set_flat( gsl_matrix * m, gsl_rng * r, double a, double b );
void  //set each element randomly, ~N(0,1)
msl_matrix_set_gaussian( gsl_matrix * m, gsl_rng * r, double sigma );
void  //set each element randomly, ~U(0,1)
msl_vector_set_flat( gsl_vector *v, gsl_rng * r, double a, double b);
void  //set each element randomly, ~N(0,1)
msl_vector_set_gaussian( gsl_vector *a, gsl_rng * r, double sigma);

int
msl_vector_equal( const gsl_vector *a, const gsl_vector *b);
int
msl_matrix_equal( const gsl_matrix *a, const gsl_matrix *b);
int
msl_vector_complex_equal( const gsl_vector_complex *a, const gsl_vector_complex *b);
int
msl_matrix_complex_equal( const gsl_matrix_complex *a, const gsl_matrix_complex *b);

void
msl_vector_pow2( gsl_vector *a, const gsl_vector *b);
void
msl_vector_abs( gsl_vector *a, const gsl_vector *b);
void
msl_vector_exp( gsl_vector *a, const gsl_vector *b);
void
msl_vector_log( gsl_vector *a, const gsl_vector *b);
void
msl_vector_sqrt( gsl_vector *a, const gsl_vector *b);

void
msl_matrix_complex_abs( gsl_matrix * , const gsl_matrix_complex *);
void
msl_matrix_complex_logabs( gsl_matrix * , const gsl_matrix_complex *);
void
msl_vector_complex_abs( gsl_vector * , const gsl_vector_complex *);
void
msl_vector_complex_logabs( gsl_vector * , const gsl_vector_complex *);


void //b_ij = abs(a_ij)
msl_matrix_abs( gsl_matrix * b, const gsl_matrix * a );
void //b_ij = log(a_ij)
msl_matrix_log( gsl_matrix * b, const gsl_matrix * a );
void //b_ij = sqrt(a_ij)
msl_matrix_sqrt( gsl_matrix * b, const gsl_matrix * a );
void //b_ij = 1.0 / a_ij
msl_matrix_scale_inv( gsl_matrix * b, const gsl_matrix * a);

void
msl_matrix_row_scal( gsl_matrix *a, const gsl_vector * v);
void
msl_matrix_row_axpy( gsl_matrix *a, const gsl_matrix * b, const gsl_vector * v);
void
msl_matrix_row_asum( gsl_vector *a, const gsl_matrix * b);
void
msl_matrix_row_nrm2( gsl_vector *a, const gsl_matrix * b);

//N is for 'Not transposed', T is for 'Transposed'
//eg. msl_matrix_gemm_NT implements z Z + alpha A B' -> a
void 
msl_matrix_gemm_NN(gsl_matrix * Z, const double z, const double alpha, const gsl_matrix *A, const gsl_matrix *B);
void
msl_matrix_gemm_NT(gsl_matrix * Z, const double z, const double alpha, const gsl_matrix *A, const gsl_matrix *B);
void 
msl_matrix_gemm_TN(gsl_matrix * Z, const double z, const double alpha, const gsl_matrix *A, const gsl_matrix *B);
void 
msl_matrix_gemm_TT(gsl_matrix * Z, const double z, const double alpha, const gsl_matrix *A, const gsl_matrix *B);

void
msl_matrix_add_column_dot( gsl_vector *v, const gsl_matrix *a, const gsl_matrix *b);
void
msl_matrix_add_column_mean( gsl_vector *m, const gsl_matrix *a );
void 
msl_matrix_add_column_variance( gsl_vector *v, const gsl_matrix *a );
void
msl_matrix_add_column_variance_m( gsl_vector *v, const gsl_matrix *a, const gsl_vector *m);

double
msl_vector_mean( gsl_vector * );
double
msl_vector_variance_m( gsl_vector *, const double mean );

void 
msl_matrix_rowmax_index( gsl_vector_ulong *amax, const gsl_matrix *a );
unsigned long int //number of rows on which the matrices take their maximum values at a common index (works when row takes max multiple times)
msl_matrix_rowmax_match( const gsl_matrix * a, const gsl_vector_ulong *amax, const gsl_matrix * b, const gsl_vector_ulong *bmax);

void   //snk.row[i] <- src.row[ rows[i] ]
msl_matrix_memcpy_rows(gsl_matrix * snk, const gsl_matrix * src, const gsl_vector_ulong *rows);

extern inline double * 
msl_matrix_view_segment( gsl_matrix_view * v, double * d, const size_t r, const size_t c)
{
    *v = gsl_matrix_view_array(d, r, c);
    return d + r * c;
}
extern inline double * 
msl_vector_view_segment( gsl_vector_view * v, double * d, const size_t s )
{
    *v = gsl_vector_view_array(d, s);
    return d + s;
}

//////////////////////////////////////////////////
//
// Functions for Neural Networks
//

extern inline double 
msl_nnet_real_logistic(double x)
{
    return 1.0 / (1.0 + exp(-x));
}
extern inline double 
msl_nnet_real_gaussian(double x, double mu, double sigma)
{
    double t = (x-mu)/sigma;
    return exp( - t * t ) / sqrt(2.0 * M_PI) / sigma;
}

//cost functions:
//w is optional weight matrix, 
//dh or dq is optional derivative of hypothesis
double 
msl_nnet_matrix_mse( const gsl_matrix *h, const gsl_matrix *y, const gsl_matrix * w, gsl_matrix * dh );
double 
msl_nnet_matrix_cross_entropy( const gsl_matrix *p, const gsl_matrix *q, const gsl_matrix *w, gsl_matrix * dq );
double 
msl_nnet_matrix_cross_entropy_1hot( const gsl_vector_ulong *p, const gsl_matrix *q, const gsl_matrix *w, gsl_matrix * dq );

//activation functions, derivatives:
void
msl_nnet_matrix_logistic( gsl_matrix * y, const gsl_matrix * x );
void
msl_nnet_matrix_logistic_set_deriv( gsl_matrix *d, const gsl_matrix *x, const gsl_matrix *y );
void
msl_nnet_matrix_logistic_add_deriv( gsl_matrix *d, const gsl_matrix *x, const gsl_matrix *y );
void
msl_nnet_matrix_softmax( gsl_matrix * y, const gsl_matrix * x );

////////////////////////////////////////////////
//
// MINIMIZATION
//
extern const gsl_multimin_fdfminimizer_type * msl_multimin_fdfminimizer_deltadelta;

double
msl_multimin_best_g_patience( 
        gsl_vector * pt, 
        gsl_multimin_function_fdf *f, 
        gsl_multimin_function *g,
        const gsl_multimin_fdfminimizer_type * algo, 
        const double algo_tol, 
        const size_t patience );

////////////////////////////////////////////////
//
// Neural Network Architectures
//

typedef struct t_msl_nnet_arch
{
    gsl_multimin_function_fdf *
        (*alloc_min)(void *, const size_t nin, const size_t nout, const size_t neg);
    void
        (*free_min)(gsl_multimin_function_fdf * );

    gsl_multimin_function *
        (*alloc_alt)(void *, const size_t nin, const size_t nout, const size_t neg);
    void
        (*free_alt)(gsl_multimin_function * );

    int 
        (*minimize)(gsl_multimin_function_fdf *, gsl_multimin_function *, gsl_vector * );

    gsl_matrix_view
        (*view_s)(void *);
    gsl_matrix_view
        (*view_y)(void *);
    gsl_matrix_view
        (*view_t)(void *);

    void
        (*reset)(void *);

    void * aparams;
} msl_nnet_arch;

msl_nnet_arch *
msl_nnet_arch_alloc_layer3_softmax(const size_t nhid);
void
msl_nnet_arch_free_layer3_softmax( msl_nnet_arch * );

msl_nnet_arch *
msl_nnet_arch_alloc_layer3_norman(const size_t nhid);
void
msl_nnet_arch_free_layer3_norman( msl_nnet_arch * );

double 
msl_nnet_arch_test_kfold( msl_nnet_arch * arch, const gsl_matrix * S, const gsl_matrix * T, const size_t nfolds, gsl_rng * r);

////////////////////////////////////////////////
//
// Functions for handling data sets
//

//a naive dictionary structure for storing a small number of strings
typedef struct t_stack_str
{
    char ** str;
    size_t size;
    size_t alloc;
} msl_stack_str;

msl_stack_str * 
msl_stack_str_alloc();
void 
msl_stack_str_free(msl_stack_str *);
void 
msl_stack_str_push(msl_stack_str *, const char * str); //O(n)
void
msl_stack_str_pop(msl_stack_str *); //O(1)
size_t 
msl_stack_str_find(const msl_stack_str *, const char * str); //O(n)

void
msl_data_onehot( gsl_matrix * , const size_t *v, const size_t size, const double hot);

typedef struct t_msl_data_labelled
{
    size_t nex;
    size_t nin;
    size_t nalloc;

    double * data_in;
    size_t * data_out;
    msl_stack_str * dict;
} msl_data_labelled;

msl_data_labelled * 
msl_data_labelled_alloc(const size_t nin);
void
msl_data_labelled_free(msl_data_labelled * mdl);
void
msl_data_labelled_realloc( msl_data_labelled * mdl, size_t cap );

msl_data_labelled * 
msl_data_Nvalue_label( const size_t N, FILE * f );

msl_data_labelled * 
msl_data_uci_iris( const char * path );
msl_data_labelled * 
msl_data_uci_ionosphere( const char * path );
msl_data_labelled * 
msl_data_uci_letter_recog( const char * path );
msl_data_labelled * 
msl_data_mnist( const char * images, const char * labels );

gsl_matrix * 
msl_data_load_spreadsheet( const char * path, const char * delims);


//////////////////////////////////////////////////
//
// Functions for digital signal processing
//
//

void
msl_dsp_vector_add_sin(gsl_vector *, double period, double phase, double amplitude);
void 
msl_dsp_matrix_estimate_harmonic_series( gsl_matrix * hs, const gsl_matrix * polarmag, const size_t fund_min, const size_t fund_max );
void
msl_dsp_matrix_lpc( gsl_matrix *lpc, gsl_matrix * ref, gsl_vector * err, gsl_matrix * ac, const gsl_matrix * signal );
void
msl_dsp_matrix_xcorr( gsl_vector *xcorr, const gsl_matrix * signal1, const gsl_matrix * signal2, const unsigned int lag );
double 
msl_dsp_vector_xcorr( const gsl_vector * signal, const gsl_vector * s2, const unsigned int lag );
void
msl_dsp_matrix_zcr( gsl_vector * , const gsl_matrix * signal );
double
msl_dsp_vector_zcr( const gsl_vector * signal );
void 
msl_dsp_matrix_energy( gsl_vector *, const gsl_matrix * signal );
void
msl_dsp_matrix_flux( gsl_vector * , const gsl_matrix * polarmag );
void
msl_dsp_matrix_hanning( gsl_matrix * a );
void
msl_dsp_vector_resample_flat( gsl_vector * a, const gsl_vector *, const size_t neighbours);
void
msl_dsp_vector_resample_triangle( gsl_vector * a, const gsl_vector *, const size_t neighbours);
void
msl_dsp_vector_resample_fn_norm( gsl_vector * a, const gsl_vector *, double (*newindex)(double));
// msl_dsp_matrix_filterbank_band()
// fills rows of @fb with triangular band-pass filters of constant area
// filter[i] is a triangle with points at freq[i], freq[i+W], freq[i+2W]
void 
msl_dsp_matrix_filterbank( gsl_matrix * fb, const gsl_vector_ulong * freq, const size_t W);
void
msl_dsp_matrix_dct( gsl_matrix * dct );
double //Mel-scale value
msl_dsp_mel_Hz(const double Hz);
double //Hz value
msl_dsp_Hz_mel(const double Mel);


gsl_vector_view
msl_dsp_vector_channel( gsl_vector * all, size_t channel, size_t nchannels);


#ifdef MSL_HAVE_LIBSNDFILE

#define MSL_RESIZE 1
#define MSL_NORESIZE 0
//if signal is alloc'ed, it is filled
//if signal is void, it is alloc'ed to be the size of the song
int //error 0 means success
msl_dsp_block_load_pcm( gsl_block * pcm, ulong * samplerate, size_t *nchannels, const char * path, const int resize );
#endif

#ifdef MSL_HAVE_MADLIB

#define MSL_RESIZE 1
#define MSL_NORESIZE 0
//if signal is alloc'ed, it is filled, starting at @offset samples into the song.
//if signal is void, it is alloc'ed to be the size of the song
int
msl_dsp_block_load_mp3( gsl_block * pcm, ulong * samplerate, size_t *nchannels, const char * path, const int resize );
#endif

#ifdef MSL_HAVE_FFTW

#include <fftw3.h>

fftw_plan
msl_dsp_matrix_row_fft( int dir, gsl_matrix_complex *, gsl_matrix *, unsigned int flags );

fftw_plan
msl_dsp_vector_fft( int dir, gsl_vector_complex *, gsl_vector *, unsigned int flags );

void 
msl_dsp_matrix_rceps( gsl_matrix *rceps, const gsl_matrix * mag );

#endif

__END_DECLS

#endif
