
/*  msl_core.c
 *
 *  Copyright (C) 2005 James Bergstra
 *
 *  This file is part of the Montreal Scientific Library.
 *
 *  The Montreal Scientific Library is free software; you can redistribute it
 *  and/or modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of the License,
 *  or (at your option) any later version.
 *
 *  The Montreal Scientific Library is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along with
 *  the Montreal Scientific Library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

/*
 *  The function for performing the Levinson-Durbin recursion for computing
 *  linear predictive coefficents (LPCs) was taken from libspeex.  It's
 *  copyright notice follows.
 */
/*
  Copyright 1992, 1993, 1994 by Jutta Degener and Carsten Bormann,
  Technische Universitaet Berlin

  Any use of this software is permitted provided that this notice is not
  removed and that neither the authors nor the Technische Universitaet Berlin
  are deemed to have made any representations as to the suitability of this
  software for any purpose nor are held responsible for any defects of
  this software.  THERE IS ABSOLUTELY NO WARRANTY FOR THIS SOFTWARE.

  As a matter of courtesy, the authors request to be informed about uses
  this software has found, about bugs in this software, and about any
  improvements that may be of general interest.

  Berlin, 28.11.1994
  Jutta Degener
  Carsten Bormann


   Code slightly modified by Jean-Marc Valin

   Speex License:

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions
   are met:
   
   - Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
   
   - Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
   
   - Neither the name of the Xiph.org Foundation nor the names of its
   contributors may be used to endorse or promote products derived from
   this software without specific prior written permission.
   
   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FOUNDATION OR
   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
   EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
   PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
   LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include "msl.h"
#include <gsl/gsl_randist.h>
#include <gsl/gsl_cblas.h>
#include <assert.h>
#include <stdint.h>
#include <string.h>

//from libspeex
/* Invented by N. Levinson in 1947, modified by J. Durbin in 1959.
 */

static inline double    /* returns minimum mean square error    */
levinson_durbin(
    double       * lpc, /*      [0...p-1] LPC coefficients      */
    const double * ac,  /*  in: [0...p] autocorrelation values  */
    double       * ref, /* out: [0...p-1] reflection coef's     */
    int p
)
        {
           int i, j;  double r, error = ac[0];

           if (ac[0] == 0) {
              for (i = 0; i < p; i++) ref[i] = 0; return 0; }

           for (i = 0; i < p; i++) {

              /* Sum up this iteration's reflection coefficient.
               */
              r = -ac[i + 1];
              for (j = 0; j < i; j++) r -= lpc[j] * ac[i - j];
              ref[i] = r /= error;   

              /*  Update LPC coefficients and total error.
               */
              lpc[i] = r;
              for (j = 0; j < i/2; j++) {
                 double tmp  = lpc[j];
                 lpc[j]     += r * lpc[i-1-j];
                 lpc[i-1-j] += r * tmp;
              }
              if (i % 2) lpc[j] += lpc[j] * r;

              error *= 1.0 - r * r;
           }
           return error;
        }
int
msl_vector_check_isnan(const gsl_vector *v)
{
    size_t i;
    for (i = 0; i < v->size; ++i)
    {
        if (!isnan( MSL_VECTOR_AT( v, i)))
        {
            return 1;
        }
    }
    return 0;
}
int
msl_vector_check_isinf(const gsl_vector *v)
{
    size_t i;
    for (i = 0; i < v->size; ++i)
    {
        if (!isinf( MSL_VECTOR_AT( v, i)))
        {
            return 1;
        }
    }
    return 0;
}

//TODO: at least make the precision #define'ed somewhere
void 
msl_vector_print_line( char ** buf, size_t *len, const gsl_vector *v)
{
    size_t i;
    size_t len_needed = 22 * v->size + 5; //more or less
    if (*len < len_needed)
    {
        *len = len_needed;
        *buf = (char *) realloc( *buf, *len );
    }
    for ( i = 0; i < v->size; ++i)
    {
        sprintf(*buf + i * 22, "%20.14e  ", MSL_VECTOR_AT(v,i));
    }
    //if ( ( strlen(*buf) != 44 ) && ( strlen(*buf) != 924 )) fprintf(stderr, "msl_vector_print_line(): strlen = %lu\n", strlen(*buf));
}
void
msl_matrix_print_spreadsheet( char **buf, size_t *len, const gsl_matrix *m )
{
    size_t i,j;
    size_t len_needed = 22 * m->size1 * m->size2 + 5; //more or less
    if (*len < len_needed)
    {
        *len = len_needed;
        *buf = (char *) realloc( *buf, *len );
    }
    for ( i = 0; i < m->size1; ++i)
    {
        for ( j = 0; j < m->size2; ++j)
        {
            sprintf(*buf + i * m->size2 * 22 + j * 22, "%20.14e %c", MSL_MATRIX_AT(m,i,j), (j == m->size2 - 1) ? '\n' : ' ');
        }
    }
}

void  //set each element randomly, ~U(0,1)
msl_matrix_set_flat( gsl_matrix * m, gsl_rng * r, double a, double b )
{
    size_t i, j;
    for ( i = 0; i < m->size1; ++i)
    {
        for (j = 0; j < m->size2; ++j)
        {
            MSL_MATRIX_AT(m,i,j) = gsl_ran_flat(r, a, b);
        }
    }
}
void  //set each element randomly, ~N(0,1)
msl_matrix_set_gaussian( gsl_matrix * m, gsl_rng * r, double sigma )
{
    size_t i, j;
    for ( i = 0; i < m->size1; ++i)
    {
        for (j = 0; j < m->size2; ++j)
        {
            MSL_MATRIX_AT(m,i,j) = gsl_ran_gaussian(r,sigma);
        }
    }
}
void  //set each element randomly, ~U(0,1)
msl_vector_set_flat( gsl_vector *v, gsl_rng * r, double a, double b)
{
    size_t i;
    for ( i = 0; i < v->size; ++i)
    {
        MSL_VECTOR_AT(v,i) = gsl_ran_flat(r, a, b);
    }
}
void  //set each element randomly, ~N(0,1)
msl_vector_set_gaussian( gsl_vector *v, gsl_rng * r, double sigma)
{
    size_t i;
    for ( i = 0; i < v->size; ++i)
    {
        MSL_VECTOR_AT(v,i) = gsl_ran_gaussian(r,sigma);
    }
}

int
msl_vector_equal( const gsl_vector *a, const gsl_vector *b)
{
    assert(a->size == b->size);
    size_t i;
    for (i = 0; i < a->size; ++i)
    {
        if (MSL_VECTOR_AT(a,i) != MSL_VECTOR_AT(b,i)) return 0;
    }
    return 1;
}
int
msl_matrix_equal( const gsl_matrix *a, const gsl_matrix *b)
{
    assert(MSL_MATRIX_SIZE_MATCH(a,b));
    size_t i,j;
    for (i = 0; i < a->size1; ++i)
    {
        for (j = 0; j < a->size2; ++j)
        {
            if (MSL_MATRIX_AT(a,i,j) != MSL_MATRIX_AT(b,i,j)) return 0;
        }
    }
    return 1;
}
int
msl_vector_complex_equal( const gsl_vector_complex *a, const gsl_vector_complex *b)
{
    assert(a->size == b->size);
    size_t i;
    for (i = 0; i < a->size; ++i)
    {
        if (MSL_VECTOR_COMPLEX_AT_R(a,i) != MSL_VECTOR_COMPLEX_AT_R(b,i)) return 0;
        if (MSL_VECTOR_COMPLEX_AT_I(a,i) != MSL_VECTOR_COMPLEX_AT_I(b,i)) return 0;
    }
    return 1;
}
int
msl_matrix_complex_equal( const gsl_matrix_complex *a, const gsl_matrix_complex *b)
{
    assert(MSL_MATRIX_SIZE_MATCH(a,b));
    size_t i,j;
    for (i = 0; i < a->size1; ++i)
    {
        for (j = 0; j < a->size2 * 2; ++j)
        {
            if (MSL_MATRIX_COMPLEX_AT_R(a,i,j) != MSL_MATRIX_COMPLEX_AT_R(b,i,j)) return 0;
            if (MSL_MATRIX_COMPLEX_AT_I(a,i,j) != MSL_MATRIX_COMPLEX_AT_I(b,i,j)) return 0;
        }
    }
    return 1;
}

//TODO: use acml_mv if available
void
msl_vector_pow2( gsl_vector *a, const gsl_vector *b)
{
    assert(a->size == b->size);
    size_t i;
    for (i = 0; i < b->size; ++i)
    {
        MSL_VECTOR_AT(a,i) = MSL_VECTOR_AT(b,i) * MSL_VECTOR_AT(b,i);
    }
}
void
msl_vector_abs( gsl_vector *a, const gsl_vector *b)
{
    assert(a->size == b->size);
    size_t i;
    for (i = 0; i < b->size; ++i)
    {
        MSL_VECTOR_AT(a,i) = fabs(MSL_VECTOR_AT(b,i));
    }
}
void
msl_vector_exp( gsl_vector *a, const gsl_vector *b)
{
    assert(a->size == b->size);
    size_t i;
    for (i = 0; i < b->size; ++i)
    {
        MSL_VECTOR_AT(a,i) = exp(MSL_VECTOR_AT(b,i));
    }
}
void
msl_vector_log( gsl_vector *a, const gsl_vector *b)
{
    assert(a->size == b->size);
    size_t i;
    for (i = 0; i < b->size; ++i)
    {
        MSL_VECTOR_AT(a,i) = log(MSL_VECTOR_AT(b,i));
    }
}
void
msl_vector_sqrt( gsl_vector *a, const gsl_vector *b)
{
    assert(a->size == b->size);
    size_t i;
    for (i = 0; i < b->size; ++i)
    {
        MSL_VECTOR_AT(a,i) = sqrt(MSL_VECTOR_AT(b,i));
    }
}

void 
msl_matrix_abs( gsl_matrix * a, const gsl_matrix * b )
{
    assert( MSL_MATRIX_SIZE_MATCH(a, b));
    size_t i, j;
    for (i = 0; i < b->size1; ++i)
    {
        for (j = 0; j < b->size2; ++j)
        {
            MSL_MATRIX_AT(a,i,j) = fabs( MSL_MATRIX_AT(b,i,j));
        }
    }
}
void 
msl_matrix_log( gsl_matrix * a, const gsl_matrix * b )
{
    assert( MSL_MATRIX_SIZE_MATCH(a, b));
    size_t i, j;
    for (i = 0; i < b->size1; ++i)
    {
        for (j = 0; j < b->size2; ++j)
        {
            MSL_MATRIX_AT(a,i,j) = log( MSL_MATRIX_AT(b,i,j));
        }
    }
}
void 
msl_matrix_sqrt( gsl_matrix * a, const gsl_matrix * b )
{
    assert( MSL_MATRIX_SIZE_MATCH(a, b));
    size_t i, j;
    for (i = 0; i < b->size1; ++i)
    {
        for (j = 0; j < b->size2; ++j)
        {
            MSL_MATRIX_AT(a,i,j) = sqrt( MSL_MATRIX_AT(b,i,j));
        }
    }
}
void //b_ij = 1.0 / a_ij, with 1.0 / 0.0 -> 0.0
msl_matrix_scale_inv( gsl_matrix * a, const gsl_matrix * b)
{
    assert( MSL_MATRIX_SIZE_MATCH(a, b));
    size_t i, j;
    for (i = 0; i < b->size1; ++i)
    {
        for (j = 0; j < b->size2; ++j)
        {
            MSL_MATRIX_AT(a,i,j) = ( MSL_MATRIX_AT(b,i,j) == 0.0)
                ? 0.0
                : 1.0 / MSL_MATRIX_AT(b,i,j);
        }
    }
}

void
msl_matrix_complex_abs( gsl_matrix *m, const gsl_matrix_complex *c)
{
    assert(MSL_MATRIX_SIZE_MATCH(m,c));
    size_t i, j;
    for (i = 0; i < m->size1; ++i)
    {
        for (j = 0; j < m->size2; ++j)
        {
            MSL_MATRIX_AT(m,i,j) = gsl_complex_abs(MSL_MATRIX_COMPLEX_AT(c,i,j));
        }
    }
}
void
msl_matrix_complex_logabs( gsl_matrix * m, const gsl_matrix_complex * c)
{
    assert(MSL_MATRIX_SIZE_MATCH(m,c));
    size_t i, j;
    for (i = 0; i < m->size1; ++i)
    {
        for (j = 0; j < m->size2; ++j)
        {
            MSL_MATRIX_AT(m,i,j) = gsl_complex_logabs(MSL_MATRIX_COMPLEX_AT(c,i,j));
        }
    }
}
void
msl_vector_complex_abs( gsl_vector *v, const gsl_vector_complex *c)
{
    assert( v->size == c->size);
    size_t j;
    for (j = 0; j < v->size; ++j)
    {
        MSL_VECTOR_AT(v,j) = gsl_complex_abs(MSL_VECTOR_COMPLEX_AT(c,j));
    }
}
void
msl_vector_complex_logabs( gsl_vector *v, const gsl_vector_complex * c)
{
    assert( v->size == c->size);
    size_t j;
    for (j = 0; j < v->size; ++j)
    {
        MSL_VECTOR_AT(v,j) = gsl_complex_logabs(MSL_VECTOR_COMPLEX_AT(c,j));
    }
}

void
msl_matrix_row_scal( gsl_matrix *a, const gsl_vector * v)
{
    assert(a->size1 == v->size);

    double * const adata = a->data;
    const int atda = a->tda;
    const int asize2 = a->size2;
    size_t i;
    for (i = 0; i < v->size; ++i)
    {
        cblas_dscal(asize2, MSL_VECTOR_AT(v,i), adata + i * atda, 1);
    }
}
void
msl_matrix_row_axpy( gsl_matrix *a, const gsl_matrix * b, const gsl_vector * v)
{
    assert(MSL_MATRIX_SIZE_MATCH(a,b));
    assert(a->size1 == v->size);

    double * const adata = a->data;
    const int atda = a->tda;
    const int asize2 = a->size2;
    size_t i;
    for (i = 0; i < v->size; ++i)
    {
        cblas_daxpy(asize2, MSL_VECTOR_AT(v,i), b->data + i * b->tda, 1, adata + i * atda, 1);
    }
}
void
msl_matrix_row_asum( gsl_vector *v, const gsl_matrix * b)
{
    assert(v->size == b->size1);

    size_t i;
    for (i = 0; i < b->size1; ++i)
    {
        MSL_VECTOR_AT(v,i) = cblas_dasum((int)b->size2, b->data + i * b->tda, 1);
    }
}
void
msl_matrix_row_nrm2( gsl_vector *v, const gsl_matrix * b)
{
    assert(v->size == b->size1);

    size_t i;
    for (i = 0; i < b->size1; ++i)
    {
        MSL_VECTOR_AT(v,i) = cblas_dnrm2((int)b->size2, b->data + i * b->tda, 1);
    }
}

void 
msl_matrix_gemm_NN(gsl_matrix * a, const double z, const double alpha, const gsl_matrix *A, const gsl_matrix *B)
{
    assert(a->size1 == A->size1);
    assert(a->size2 == B->size2);
    assert(A->size2 == B->size1);
    //if (A->tda == 0) assert(0);
    //fprintf(stderr, "tdas: %lu %lu %lu\n", A->tda, B->tda, tda);
    cblas_dgemm (CblasRowMajor, 
         CblasNoTrans, CblasNoTrans, (int)a->size1, (int)a->size2, (int)A->size2,
         alpha, A->data, (int)A->tda, B->data, (int)B->tda, z, a->data, (int)a->tda);
}
void
msl_matrix_gemm_NT(gsl_matrix * a, const double z, const double alpha, const gsl_matrix *A, const gsl_matrix *B)
{
    assert(a->size1 == A->size1);
    assert(a->size2 == B->size1);
    assert(A->size2 == B->size2);
    cblas_dgemm (CblasRowMajor, 
         CblasNoTrans, CblasTrans, (int)a->size1, (int)a->size2, (int)A->size2,
         alpha, A->data, (int)A->tda, B->data, (int)B->tda, z, a->data, (int)a->tda);
}
void 
msl_matrix_gemm_TN(gsl_matrix * a, const double z, const double alpha, const gsl_matrix *A, const gsl_matrix *B)
{
    assert(a->size1 == A->size2);
    assert(a->size2 == B->size2);
    assert(A->size1 == B->size1);
    cblas_dgemm (CblasRowMajor, 
         CblasTrans, CblasNoTrans, (int)a->size1, (int)a->size2, (int)A->size1,
         alpha, A->data, (int)A->tda, B->data, (int)B->tda, z, a->data, (int)a->tda);
}
void 
msl_matrix_gemm_TT(gsl_matrix * a, const double z, const double alpha, const gsl_matrix *A, const gsl_matrix *B)
{
    assert(a->size1 == A->size2);
    assert(a->size2 == B->size1);
    assert(A->size1 == B->size2);
    cblas_dgemm (CblasRowMajor, 
         CblasTrans, CblasTrans, (int)a->size1, (int)a->size2, (int)A->size1,
         alpha, A->data, (int)A->tda, B->data, (int)B->tda, z, a->data, (int)a->tda);
}

void
msl_matrix_add_column_dot( gsl_vector *v, const gsl_matrix *a, const gsl_matrix *b)
{
    size_t i;
    assert(MSL_MATRIX_SIZE_MATCH(a,b));
    assert(v->size == a->size2);
    for (i = 0; i < v->size; ++i)
    {
        MSL_VECTOR_AT(v,i) = cblas_ddot((int)a->size1, a->data + i, (int)a->tda, b->data + i, (int)b->tda );
    }
}
void
msl_matrix_add_column_mean( gsl_vector *v, const gsl_matrix *a )
{
    size_t i;
    assert(v->size == a->size2);
    assert(a->size1);
    for (i = 0; i < a->size1; ++i) //for each row
    {
        cblas_daxpy((int)v->size, 1.0, a->data + i * a->tda, 1, v->data, (int)v->stride );
    }
    cblas_dscal((int)v->size, 1.0 / a->size1, v->data, (int)v->stride);
}
void 
msl_matrix_add_column_variance( gsl_vector *v, const gsl_matrix *a )
{
    gsl_vector *m = gsl_vector_alloc(v->size);
    gsl_vector_set_zero(m);
    msl_matrix_add_column_mean(m,a);
    msl_matrix_add_column_variance_m(v,a,m);
    gsl_vector_free(m);
}
void
msl_matrix_add_column_variance_m( gsl_vector *v, const gsl_matrix *a, const gsl_vector *m)
{
    assert(v->size == a->size2);
    assert(v->size == m->size);
    gsl_vector *d = gsl_vector_alloc(v->size);
    size_t i;
    for (i = 0; i < a->size1; ++i) //foreach row
    {
        cblas_dcopy( (int)d->size, m->data, (int)m->stride, d->data, 1); //d = m
        cblas_daxpy( (int)d->size, -1.0, a->data + i * a->tda, 1, d->data, 1); //d = m - a_i
        gsl_vector_mul(d, d); //d = (m - a_i)^2
        cblas_daxpy( (int)d->size, 1.0, d->data, 1, v->data, (int)v->stride);//v += (m - a_i)^2
    }
    if (a->size1 > 1)
    {
        cblas_dscal((int)v->size, 1.0 / (a->size1-1), v->data, (int)v->stride);
    }
    gsl_vector_free(d);
}

double
msl_vector_mean( gsl_vector * v)
{
    return (v->size) ? cblas_dasum( (int)v->size, v->data, (int)v->stride) / (double) v->size : 0.0;
}
double
msl_vector_variance_m( gsl_vector * v , const double mean )
{
    double rval = 0;
    size_t i;
    for (i = 0; i < v->size; ++i)
    {
        rval += ( MSL_VECTOR_AT(v,i) - mean)*( MSL_VECTOR_AT(v,i) - mean);
    }
    return rval;
}

void 
msl_matrix_rowmax_index( gsl_vector_ulong *amax, const gsl_matrix *a )
{
    assert(amax->size == a->size1);
    size_t i;
    for (i = 0; i < amax->size; ++i)
    {
        MSL_VECTOR_AT(amax,i) = cblas_idamax((int)a->size2, a->data + i * a->tda, 1);
    }
}
size_t //number of rows on which the matrices take their maximum values at a common index (works when row takes max multiple times)
msl_matrix_rowmax_match( const gsl_matrix * a, const gsl_vector_ulong *amax, const gsl_matrix * b, const gsl_vector_ulong *bmax)
{
    assert( MSL_MATRIX_SIZE_MATCH(a, b));
    assert(amax->size == a->size1);
    assert(bmax->size == b->size1);
    size_t rval = 0;
    size_t i;
    for (i = 0; i < a->size1; ++i)
    {
        rval += (  ( MSL_MATRIX_AT(a, i, MSL_VECTOR_AT(amax,i)) == MSL_MATRIX_AT(a, i, MSL_VECTOR_AT(bmax,i)))
                || ( MSL_MATRIX_AT(b, i, MSL_VECTOR_AT(amax,i)) == MSL_MATRIX_AT(b, i, MSL_VECTOR_AT(bmax,i ))))
            ? 1 : 0;
    }
    return rval;
}
void
msl_matrix_memcpy_rows(gsl_matrix * snk, const gsl_matrix * src, const gsl_vector_ulong *rows)
{
    assert( snk->size2 == src->size2);
    assert( snk->size1 == rows->size);
    size_t i;
    for (i = 0; i < rows->size; ++i)
    {
        assert( MSL_VECTOR_AT(rows, i) < src->size1 );
        memcpy(snk->data + i * snk->tda, src->data + MSL_VECTOR_AT(rows, i) * src->tda, src->size2 * sizeof(double));
    }
}

double * 
msl_matrix_view_segment( gsl_matrix_view * v, double * d, const size_t r, const size_t c)
{
    *v = gsl_matrix_view_array(d, r, c);
    return d + r * c;
}
double * 
msl_vector_view_segment( gsl_vector_view * v, double * d, const size_t s )
{
    *v = gsl_vector_view_array(d, s);
    return d + s;
}

//does a minimization of f
//and checks at each iteration the value of g
//the function returns the best value observed on g, and the point at which it
//was attained
double
msl_multimin_best_g_patience( 
        gsl_vector * pt, 
        gsl_multimin_function_fdf *f, 
        gsl_multimin_function *g,
        const gsl_multimin_fdfminimizer_type * algo, 
        const double algo_tol, 
        const size_t patience )
{
    double best_g = GSL_POSINF;
    const double firststep = 0.001;
    assert(f->n == g->n);
    gsl_multimin_fdfminimizer * mizer = gsl_multimin_fdfminimizer_alloc (algo, f->n);
    gsl_multimin_fdfminimizer_set( mizer, f, pt, firststep, algo_tol);
    //set up a multimin for f.
    size_t ticks = 0;
    while ((0 == gsl_multimin_fdfminimizer_iterate(mizer)) && ( ticks < patience ) )
    {
        double this_g = GSL_MULTIMIN_FN_EVAL(g,mizer->x);
//        fprintf(stderr, "best_g() updating weights %lf %lf\n", this_g, best_g );
        if ( this_g < best_g )
        {
            gsl_vector_memcpy( pt, mizer->x );
            best_g = this_g;
            ticks = 0;
        }
        else
        {
            ++ticks;
        }
    }

    gsl_multimin_fdfminimizer_free(mizer);

    return best_g;
}


double msl_nnet_real_logistic(double x)
{
    return 1.0 / (1.0 + exp(-x));
}
double msl_nnet_real_gaussian(double x, double mu, double sigma)
{
    double t = (x-mu)/sigma;
    return exp( - t * t ) / sqrt(2.0 * M_PI) / sigma;
}

double  //sum_ij 1/2 w_ij (y_ij - h_ij)^2   , d_ij = w_ij(h_ij - y_ij)
msl_nnet_matrix_mse( const gsl_matrix *h, const gsl_matrix *y, const gsl_matrix *w, gsl_matrix *dh  )
{
    assert(MSL_MATRIX_SIZE_MATCH(h,y));
    if (w)
    {
        assert(MSL_MATRIX_SIZE_MATCH(w,y));
    }
    if (dh)
    {
        assert(MSL_MATRIX_SIZE_MATCH(dh,y));
    }

    double rval = 0;
    double _dh;
    size_t i,j;
    for ( i = 0; i < h->size1; ++i)
    {
        for (j = 0; j < h->size2; ++j)
        {
            _dh = MSL_MATRIX_AT(h,i,j) - MSL_MATRIX_AT(y,i,j);
            if (dh) MSL_MATRIX_AT(dh,i,j) = _dh * (w ? MSL_MATRIX_AT(w,i,j) : 1.0);
            rval += _dh * _dh * (w ? MSL_MATRIX_AT(w,i,j) : 1.0);
        }
    }
    return rval * 0.5;
}
double 
msl_nnet_matrix_cross_entropy( const gsl_matrix *p, const gsl_matrix *q, const gsl_matrix *w, gsl_matrix *dq )
{
    assert(MSL_MATRIX_SIZE_MATCH(p,q));
    if (w)
    {
        assert(MSL_MATRIX_SIZE_MATCH(w,p));
    }
    if (dq)
    {
        assert(MSL_MATRIX_SIZE_MATCH(dq,p));
    }
    size_t i, j;
    double rval = 0.0;
    for (i = 0; i < p->size1; ++i) 
    {
        for (j = 0; j < p->size2; ++j)
        {
            rval -= ( MSL_MATRIX_AT(p,i,j) > 0.0 )
                ? MSL_MATRIX_AT(p,i,j) * log(MSL_MATRIX_AT(q,i,j)) * (w ? MSL_MATRIX_AT(w,i,j):1.0)
                : 0.0;
            if (dq)
            {
                MSL_MATRIX_AT(dq,i,j) = (MSL_MATRIX_AT(q,i,j) - MSL_MATRIX_AT(p,i,j)) * (w ? MSL_MATRIX_AT(w,i,j):1.0);
            }
        }
    }
    return rval;
}
double 
msl_nnet_matrix_cross_entropy_1hot( const gsl_vector_ulong *p, const gsl_matrix *q, const gsl_matrix * w, gsl_matrix *dq )
{
    assert(p->size == q->size1);
    assert(0);
}

void    //b_ij = 1.0 / (1.0 + exp(-a_ij))
msl_nnet_matrix_logistic( gsl_matrix * a, const gsl_matrix * b )
{
    assert(MSL_MATRIX_SIZE_MATCH(a,b));

    size_t i,j;
    for ( i = 0; i < b->size1; ++i)
    {
        for (j = 0; j < b->size2; ++j)
        {
            MSL_MATRIX_AT(a,i,j) = msl_nnet_real_logistic(MSL_MATRIX_AT(b,i,j));
        }
    }
}
void
msl_nnet_matrix_logistic_set_deriv( gsl_matrix *ddx, const gsl_matrix *ddy, const gsl_matrix *y )
{
    assert(MSL_MATRIX_SIZE_MATCH(ddx,y));
    assert(MSL_MATRIX_SIZE_MATCH(ddy,y));

    size_t i,j;
    for ( i = 0; i < y->size1; ++i)
    {
        for (j = 0; j < y->size2; ++j)
        {
            MSL_MATRIX_AT(ddx,i,j) = MSL_MATRIX_AT(ddy,i,j) * MSL_MATRIX_AT(y,i,j) * (1.0 - MSL_MATRIX_AT(y,i,j));
        }
    }
}
void
msl_nnet_matrix_logistic_add_deriv( gsl_matrix *ddx, const gsl_matrix *ddy, const gsl_matrix *y )
{
    assert(MSL_MATRIX_SIZE_MATCH(ddx,y));
    assert(MSL_MATRIX_SIZE_MATCH(ddy,y));

    size_t i,j;
    for ( i = 0; i < y->size1; ++i)
    {
        for (j = 0; j < y->size2; ++j)
        {
            MSL_MATRIX_AT(ddx,i,j) += MSL_MATRIX_AT(ddy,i,j) * MSL_MATRIX_AT(y,i,j) * (1.0 - MSL_MATRIX_AT(y,i,j));
        }
    }
}

void    //b_ij = exp(a_ij) / sum_k exp(a_ik)
msl_nnet_matrix_softmax( gsl_matrix * a, const gsl_matrix * b )
{
    assert(MSL_MATRIX_SIZE_MATCH(a,b));

    double rsum;
    size_t i,j;
    for ( i = 0; i < b->size1; ++i)
    {
        rsum = 0;
        for (j = 0; j < b->size2; ++j)
        {
            rsum += MSL_MATRIX_AT(a,i,j) = exp(MSL_MATRIX_AT(b,i,j));
        }
        cblas_dscal( (int)a->size2, 1.0 / rsum, a->data + i * a->tda, 1);
    }
}

void
msl_dsp_vector_add_sin(gsl_vector *v, double period, double phase, double amplitude)
{
    double a = phase;
    const double da = 2 * M_PI / period;
    size_t i;
    for (i = 0; i < v->size; ++i)
    {
        MSL_VECTOR_AT(v,i) += amplitude * sin( a );
        a += da;
    }
}
void 
msl_dsp_matrix_estimate_harmonic_series( gsl_matrix * hs, const gsl_matrix * polarmag, const size_t fund_min, const size_t fund_max )
{
    hs = hs;
    polarmag = polarmag;
    assert(0);
}
void
msl_dsp_matrix_lpc( gsl_matrix *lpc, gsl_matrix * ref, gsl_vector * rec_err, gsl_matrix * ac, const gsl_matrix * signal )
{
    //number of lpc's computed comes from size2 of lpc
    //number of autocorrelation coeffs computed @ac is lpc->size2 + 1
    assert(MSL_MATRIX_SIZE_MATCH( lpc, ref ));
    assert(lpc->size1 == signal->size1);
    assert(lpc->size1 == rec_err->size);
    assert(ac->size1 == lpc->size1);
    assert(ac->size2 == lpc->size2 + 1);

    size_t i;
    gsl_vector_view vv;

    for (i = 0; i < ac->size2; ++i) //ac[r,c] = autocorr(signal_r, lag=c)
    {
        vv = gsl_matrix_column(ac, i);
        msl_dsp_matrix_xcorr( &vv.vector, signal, signal, (uint)i );
    }

    for (i = 0; i < ac->size1; ++i ) //rec_err_r = lev_durb(lpc,ac,ref,nlpc)
    {
        MSL_VECTOR_AT(rec_err, i) = levinson_durbin( 
                lpc->data + i * lpc->tda, 
                ac->data + i * ac->tda, 
                ref->data + i * ref->tda,
                (int)lpc->size2 );
    }

}
void
msl_dsp_matrix_xcorr( gsl_vector *xcorr, const gsl_matrix * signal1, const gsl_matrix * signal2, const uint lag )
{
    assert(xcorr->size == signal1->size1);
    assert(xcorr->size == signal2->size1);

    size_t i;
    const int N = GSL_MIN( signal1->size2, signal2->size2 - lag);
    for (i = 0; i < xcorr->size; ++i)
    {
        MSL_VECTOR_AT(xcorr,i) =  cblas_ddot( N, signal1->data + i * signal1->tda, 1, signal2->data + i * signal2->tda + lag, 1 );
    }
}
void
msl_dsp_matrix_zcr( gsl_vector *v , const gsl_matrix * signal )
{
    size_t  i,j;
    for (i = 0; i < signal->size1; ++i)
    {
        MSL_VECTOR_AT(v,i) = 0;
        for (j = 1; j < signal->size2; ++j)
        {
            MSL_VECTOR_AT(v,i) += ( MSL_MATRIX_AT(signal, i,j) * MSL_MATRIX_AT(signal, i, j-1) > 0 ) ? 0.0 : 1.0;
        }
    }
    if (signal->size2 > 1)
    {
        cblas_dscal((int)v->size, 1.0 / signal->size2, v->data, (int)v->stride);
    }
}
void 
msl_dsp_matrix_energy( gsl_vector *v, const gsl_matrix * signal )
{
    size_t  i;
    for (i = 0; i < signal->size1; ++i)
    {
        MSL_VECTOR_AT(v,i) = cblas_ddot( (int)signal->size2, signal->data + i * signal->tda, 1, signal->data + i * signal->tda, 1);
    }
    if (signal->size2 > 1)
    {
        cblas_dscal((int)v->size, 1.0 / signal->size2, v->data, (int)v->stride);
    }
}
void
msl_dsp_matrix_centroid( gsl_vector *v , const gsl_matrix * polarmag ){assert(0);}
void
msl_dsp_matrix_spectral_rolloff( gsl_vector * v, const gsl_matrix * polarmag ){assert(0);}
void
msl_dsp_matrix_spectral_flux( gsl_vector * v, const gsl_matrix * polarmag ){assert(0);}
void
msl_dsp_matrix_hanning( gsl_matrix * a ){assert(0);}
void
msl_dsp_matrix_resample_lin( gsl_matrix * a, const gsl_matrix * m, const size_t neighbours, const int mode){assert(0);}
void
msl_dsp_matrix_resample_nonlin( gsl_matrix * a, const gsl_matrix * m, double (*newindex)(double)){assert(0);}
void
msl_dsp_vector_resample_neighb_flat( gsl_vector * a, const gsl_vector *b, const size_t neighbours)
{
    const double ratio = (double) b->size / (double) a->size;
    int i;
    int j;
    const int I = a->size;
    const int w = neighbours / 2;

    gsl_vector_set_zero(a);

    for (i = 0; i < I; ++i)
    {
        for (j = -w; j < w + 1; ++j)
        {
            const int idx = i * ratio + w;
            if (idx >= 0 && (unsigned)idx < b->size)
            {
                MSL_VECTOR_AT(a,i) += MSL_VECTOR_AT(b,idx);
            }
        }
    }
}
void
msl_dsp_vector_resample_neighb_triangle( gsl_vector * a, const gsl_vector *b, const size_t neighbours)
{
    const double ratio = (double) b->size / (double) a->size;
    int i;
    int j;
    const int I = a->size;
    const int w = neighbours / 2;

    gsl_vector_set_zero(a);

    for (i = 0; i < I; ++i)
    {
        for (j = -w; j < w + 1; ++j)
        {
            const int idx = i * ratio + w;
            if (idx >= 0 && (unsigned) idx < b->size)
            {
                MSL_VECTOR_AT(a,i) += MSL_VECTOR_AT(b,idx) * (w + 1 - abs(j));
            }
        }
    }
}
void
msl_dsp_vector_resample_fn_norm( gsl_vector * a, const gsl_vector * b, double (*newindex)(double))
{
    size_t i;

    gsl_vector_set_zero(a);

    for (i = 0; i < b->size; ++i)
    {
        double ni = newindex((double)i / (double)b->size) * a->size;
        const double nmod = fmod(ni, 1.0);
        ni -= nmod;
        assert(ni < (double) a->size);
        assert(ni >= 0.0 );
        MSL_VECTOR_AT( a, (size_t) ni ) += (1.0 - nmod) * MSL_VECTOR_AT(b,i);
        if ( ((size_t)ni + 1 < a->size ) && (i + 1 < b->size))
        {
            MSL_VECTOR_AT( a, (size_t)ni + 1 ) += nmod * MSL_VECTOR_AT(b,i);
        }
    }
}

double
msl_dsp_vector_zcr( const gsl_vector * signal )
{
    size_t  i;
    double rval = 0.0;
    for (i = 1; i < signal->size; ++i)
    {
        rval += ( MSL_VECTOR_AT(signal, i) * MSL_VECTOR_AT(signal, i-1) > 0 ) ? 0.0 : 1.0;
    }
    rval /= signal->size;
    return rval;
}
double 
msl_dsp_vector_xcorr( const gsl_vector *signal1, const gsl_vector *signal2, const uint lag )
{
    return cblas_ddot(
            (signal1->size < signal2->size - lag) ? (int)signal1->size: (int)signal2->size - (int)lag,
            signal1->data, (int)signal1->stride,
            signal2->data + lag * signal2->stride, (int)signal2->stride);
}
void
msl_dsp_matrix_filterbank( gsl_matrix * fb, const gsl_vector_ulong * freq, const size_t W)
{
    size_t i,j;
    assert(fb->size1 == freq->size + 2 * W);
    for (i = 0; i < fb->size1; ++i)
    {
        for (j = 0; j < fb->size2; ++j)
        {
            MSL_MATRIX_AT( fb, i, j ) = 2.0 / (double)(MSL_VECTOR_AT(freq,i+2*W) - MSL_VECTOR_AT(freq,i)) * 
                ( j < MSL_VECTOR_AT(freq,i) )
                ? 0.0
                : ( j > MSL_VECTOR_AT( freq, i + 2 * W))
                ? 0.0
                : ( j < MSL_VECTOR_AT( freq, i + 1 * W))
                ? (double)( j - MSL_VECTOR_AT( freq, i ) ) / (double)( MSL_VECTOR_AT(freq, i+1*W) - MSL_VECTOR_AT(freq,i) )
                : (double)( MSL_VECTOR_AT(freq,i+2*W) - j) / (double)( MSL_VECTOR_AT(freq, i+2*W) - MSL_VECTOR_AT(freq,i+1*W) ) ;
        }
    }
}
void
msl_dsp_matrix_dct( gsl_matrix * dct )
{
    double d = 1.0 / sqrt(dct->size2);
    size_t i,j;
    for (i = 0; i < dct->size1; ++i) //for each cepstral coef
    {
        for (j = 0; j < dct->size2; ++j) //for each filter
        {
            MSL_MATRIX_AT(dct, i, j) = (i == 0)
                ? d * cos(i * (2*j+1) * M_PI_2/dct->size2) * sqrt(2.0)/2.0 //the additional coefficient is Tzanetakis's special touch
                : d * cos(i * (2*j+1) * M_PI_2/dct->size2);
        }
    }
}

double 
msl_dsp_mel_Hz(const double x)
{
    return 2595.0 * log10(1.0 + x / 700.0);
}
double
msl_dsp_Hz_mel(const double x)
{
    return 700.0 * ( pow(10.0, x/2595.0) - 1.0 );
}

#if 0
void
msl_dsp_vector_ulong_melinv( gsl_vector_ulong * freq, const double bot_freq, const double top_freq) 
{
    size_t i,j;
    const size_t n_exp = fb->size1 - n_linear;
    const size_t N = fb->size1;
    double expfac = 1;
    gsl_vector * freq = gsl_vector_alloc( N + 2 );
    gsl_vector * bmag = gsl_vector_alloc( N );

    // init frequency boundaries
    // ... linear portion
    for (i = 0; i < n_linear; ++i)
    {
        MSL_VECTOR_AT(freq,i) = lowest_freq + i * linstep;
    }
    // ... exponential portion
    for (i = 0; i < n_exp + 2; ++i)
    {
        expfac *= expstep;
        MSL_VECTOR_AT(freq, n_linear + i) = expfac * MSL_VECTOR_AT(freq, n_linear-1);
    }
    assert( MSL_VECTOR_AT(freq, N) < (double) fb->size2 );

    for (i = 0; i < fb->size1; ++i)
    {
        for (j = 0; j < fb->size2; ++j)
        {
            MSL_MATRIX_AT(fb, i, j) = 
                ( (double) j < MSL_VECTOR_AT(freq,i) )
                ? 0
                : ((double) j > MSL_VECTOR_AT(freq,i+2))
                ? 0
                : ((double) j < MSL_VECTOR_AT(freq,i + 1))
                ? ( (double) j - MSL_VECTOR_AT(freq,i) ) / ( MSL_VECTOR_AT(freq, i+1) - MSL_VECTOR_AT(freq,i) )
                : ( MSL_VECTOR_AT(freq,i+2) - (double) j)  / ( MSL_VECTOR_AT(freq, i+2) - MSL_VECTOR_AT(freq,i+1) ) ;
        }
    }

    //set up triangle widths
    for (i = 0; i < fb->size1; ++i)
    {
        MSL_VECTOR_AT(bmag,i) = 2.0 / ( MSL_VECTOR_AT(freq, i + 2) - MSL_VECTOR_AT(i) );
    }

    //scale the filterbank so that each filter is a triangle of constant area
    msl_matrix_row_scal(fb, bmag);

    gsl_vector_free(bmag);
    gsl_vector_free(freq);
}
#endif

gsl_vector_view
msl_dsp_vector_channel(gsl_vector * all, size_t channel, size_t nchannels)
{
    assert(channel < nchannels);
    return gsl_vector_subvector_with_stride(all, channel, nchannels, all->size / nchannels);
}


