
/*  msl_data.c
 *
 *  Copyright (C) 2005 James Bergstra
 *
 *  This file is part of the Montreal Scientific Library.
 *
 *  The Montreal Scientific Library is free software; you can redistribute it
 *  and/or modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of the License,
 *  or (at your option) any later version.
 *
 *  The Montreal Scientific Library is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along with
 *  the Montreal Scientific Library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#define _GNU_SOURCE
#include "msl.h"
#include <string.h>
#include <assert.h>

#define DECL_MALLOC_PTR(t,v) t * v = (t*)malloc(sizeof(t))
#define DECL_CAST(t,v,p) t * v = (t *)( p );

msl_stack_str * 
msl_stack_str_alloc()
{
    DECL_MALLOC_PTR(msl_stack_str, r);    
    r->str = (char **) malloc(10 * sizeof(char*));
    r->size = 1;
    r->alloc = 10;
    r->str[0] = NULL; // 0 is a negative find result
    return r;
}
void 
msl_stack_str_free(msl_stack_str *r)
{
    size_t i;
    for (i = 1; i < r->size; ++i) free(r->str[i]);
    free(r->str);
    free(r);
}
void 
msl_stack_str_push(msl_stack_str *r, const char * str)
{
    if (r->size == r->alloc){
        r->alloc *= 2;
        realloc(&r->str, r->alloc);
    }
    r->str[ r->size++ ] = strdup( str );
}
void 
msl_stack_str_pop(msl_stack_str *r)
{
    if (r->size > 1)
    {
        free(r->str[ --r->size ] );
    }
}
size_t 
msl_stack_str_find(const msl_stack_str *r, const char * str)
{
    size_t i = r->size - 1;
    while ( ! ((i == 0) || (0 == strcmp( str, r->str[i] )))) --i;
    return i;
}
void
msl_data_onehot( gsl_matrix *m, const size_t *v, const size_t size, const double hot)
{
    assert(m->size1 == size);
    size_t i;
    gsl_matrix_set_zero(m);
    for (i = 0; i < size; ++i)
    {
        assert(v[i] < m->size2);
        MSL_MATRIX_AT(m,i, v[i]) = hot;
    }
}

msl_data_labelled * 
msl_data_labelled_alloc(const size_t nin)
{
    DECL_MALLOC_PTR(msl_data_labelled, mdl);
    mdl->nex = 0;
    mdl->nin = nin;
    mdl->nalloc = 0;
    mdl->data_in = NULL;
    mdl->data_out = NULL;
    mdl->dict = msl_stack_str_alloc();
    return mdl;
}
void
msl_data_labelled_free(msl_data_labelled * mdl)
{
    fprintf(stderr, "msl_data_labelled_free() not implemented\n");
}
void
msl_data_labelled_realloc( msl_data_labelled * mdl, size_t nalloc )
{
    mdl->nalloc = nalloc;
    mdl->data_in = realloc(mdl->data_in, nalloc * mdl->nin * sizeof(double));
    mdl->data_out = realloc(mdl->data_out, nalloc * sizeof(ulong));

}

msl_data_labelled * 
msl_data_uci_iris( const char * path )
{
    msl_data_labelled * mdl = msl_data_labelled_alloc( 4l );
    msl_data_labelled_realloc(mdl, 32l);

    FILE * f = fopen( "iris.data", "r" );
    char buf[128];

    while (!feof(f))
    {
        if (mdl->nex == mdl->nalloc) msl_data_labelled_realloc(mdl, mdl->nalloc * 2);

        double * p = mdl->data_in + mdl->nex * mdl->nin;
        fscanf(f,"%lf,%lf,%lf,%lf,%s\n", p+ 0, p+ 1, p+ 2, p+ 3, buf);
        mdl->data_out[mdl->nex] = msl_stack_str_find(mdl->dict, buf);
        if (0 == mdl->data_out[mdl->nex]) 
        {
            mdl->data_out[mdl->nex] = mdl->dict->size;
            msl_stack_str_push(mdl->dict, buf);
        }
        ++mdl->nex;
    }
    fclose(f);

    return mdl;
}

msl_data_labelled * 
msl_data_Nvalue_label( const size_t N, FILE * f )
{
    msl_data_labelled * mdl = msl_data_labelled_alloc( N );
    msl_data_labelled_realloc(mdl, 32l);

    char buf[128];
    size_t i;

    while (1)
    {
        if (mdl->nex == mdl->nalloc) msl_data_labelled_realloc(mdl, mdl->nalloc * 2);

        double * p = mdl->data_in + mdl->nex * mdl->nin;
        for (i = 0; i < N; ++i) {
            if (1 != fscanf(f,"%lf,", p)) 
            {
                if (i) fprintf(stderr, "msl_data_Nvalue_label() Failed mid-row to read floating point value :( \n");
                break;
            }
            ++p;
        }

        if (i < N) break;

        if (1 != fscanf(f, "%s\n", buf))
        {
            fprintf(stderr, "msl_data_Nvalue_label() Failed to read trailing label.\n");
            exit(1);
        }
        mdl->data_out[mdl->nex] = msl_stack_str_find(mdl->dict, buf);
        if (0 == mdl->data_out[mdl->nex]) 
        {
            mdl->data_out[mdl->nex] = mdl->dict->size;
            msl_stack_str_push(mdl->dict, buf);
        }
        ++mdl->nex;
    }

    return mdl;
}

#if 0
msl_data_labelled *
msl_data_uci_ionosphere( const char * path )
{
    DECL_MALLOC_PTR(msl_data_labelled, r);    
        FILE * f = fopen( "ionosphere.data", "r" );
        char c;

        while (!feof(f)) {
            push_back(Example(2));
            back()[0].alloc(1,34);
            back()[1].alloc(1,2);
            back()[1].zero();
            for (size_t i = 0; i < 34; ++i) {
                fscanf(f, "%lf,", &back()[0].at(0,i));
            //    fprintf(stdout, "%lf..\t", back()[0].at(0,i));
            }
            fscanf(f, "%c\n", &c );
            if (c == 'g') back()[1].at(0,0) = 1.0;
            else if (c == 'b') back()[1].at(0,1) = 1.0;
            else assert(0);

//            back()[0].print(stdout);
//            back()[1].print(stdout);
//            exit(1);
        }
        fprintf(stderr, "UCI IonoSphere: read %lu samples\n", size());
        fclose(f);
}
msl_data_labelled *
msl_data_uci_spambase( const char * path )
{
    public:
    TaskSpamBase()
    {
        FILE * f = fopen( "spambase.data", "r" );
        char c;

        while (!feof(f)) {
            push_back(Example(2));
            back()[0].alloc(1,57);
            back()[1].alloc(1,2);
            back()[1].zero();
            for (size_t i = 0; i < back()[0].size2(); ++i) {
                fscanf(f, "%lf,", &back()[0].at(0,i));
            //    fprintf(stdout, "%lf..\t", back()[0].at(0,i));
            }
            fscanf(f, "%c\n", &c );
            if (c == '0') back()[1].at(0,0) = 1.0;
            else if (c == '1') back()[1].at(0,1) = 1.0;
            else assert(0);

           // back()[0].print(stdout);
           // back()[1].print(stdout);
           // exit(1);
        }
        fprintf(stderr, "UCI Spambase: read %lu samples\n", size());
        fclose(f);
    }
};
msl_data_labelled * 
msl_data_uci_letters( const char * path)
{
    TaskLetterRecognition()
    {
        FILE * f = fopen( "letter-recognition.data", "r" );
        char c;

        while (!feof(f)) {
            push_back(Example(2));
            back()[0].alloc(1,16);
            back()[1].alloc(1,26);
            back()[1].zero();
            fscanf(f, "%c", &c );
            back()[1].at(0,c-'A') = 1.0;
            for (size_t i = 0; i < back()[0].size2(); ++i) {
                fscanf(f, ",%lf", &back()[0].at(0,i));
            //    fprintf(stdout, "%lf..\t", back()[0].at(0,i));
            }
            fscanf(f, "\n");

            //back()[0].print(stdout);
            //back()[1].print(stdout);
            //exit(1);
        }
        fprintf(stderr, "UCI Spambase: read %lu samples\n", size());
        fclose(f);
    }
};

#endif

gsl_matrix *
msl_data_load_spreadsheet( const char * path, const char * delims )
{
    char *line = NULL;
    char *line1, *line2;
    size_t linelen = 0;
    size_t linenum = 0;
    size_t size1 = 0;
    size_t size2 = 0;
    size_t size2_line = 0;
    double *data = 0;
    size_t data_len = 0;
    size_t data_alloc = 0;

    FILE * f = fopen(path, "r");
    if (NULL == f)
    {
        return NULL;
    }

    while (0 < getline( &line, &linelen, f ) ) 
    {
        ++linenum;
        line1 = line;
        line2 = NULL;
        size2_line = 0;
        while (1) 
        {
            if ( data_alloc == data_len )
            {
                data_alloc = (data_alloc == 0)? 256: data_alloc * 2;
                data = (double *)realloc(data, data_alloc * sizeof(double));
            }
            double d = strtod( line1, &line2 );
            if (line1 == line2 ) break; //either error or end of line, who cares... 
            //if it was an error we'll catch it if it makes
            //a line-length mismatch 
            //NB: this allows arbitrary non-numerical text
            //to follow the matrix
            else data[ data_len++ ] = d;
            ++size2_line;
            line1 = line2;
            while (strchr( delims, *line1)) ++line1;
        } 

        if (size2_line > 0)
        {
            if (size1 == 0) size2 = size2_line;
            else if (size2 != size2_line) 
            {
                fprintf(stderr, "ERROR: line=%lu size2=%lu size2_line=%lu\n", linenum, size2, size2_line);
                exit(1);
            }
            ++size1;
        }
    }

    fclose(f);
    //now the initialization of parameters:
    //now switch from a malloc()'ed buffer to a new'ed one
    gsl_matrix * m = gsl_matrix_alloc(size1, size2);
    memcpy( m->data, data, size1 * size2 * sizeof(size_t));
    free(data);
    return m;
}

