
/*  msl_deltadelta.c
 *
 *  Copyright (C) 2005 James Bergstra
 *
 *  This file is part of the Montreal Scientific Library.
 *
 *  The Montreal Scientific Library is free software; you can redistribute it
 *  and/or modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of the License,
 *  or (at your option) any later version.
 *
 *  The Montreal Scientific Library is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along with
 *  the Montreal Scientific Library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "msl.h"


typedef struct delta_delta_state_t
{
    double step;
    double default_step;
    double tol;
    gsl_vector *x1;
    gsl_vector *g1;
    gsl_vector *lr;
} delta_delta_state_t;

static int delta_delta_alloc (void *vstate, size_t n)
{
  delta_delta_state_t *state = (delta_delta_state_t *) vstate;

  state->x1 = gsl_vector_alloc (n);

  if (state->x1 == NULL)
    {
      GSL_ERROR ("failed to allocate space for x1", GSL_ENOMEM);
    }

  state->g1 = gsl_vector_alloc (n);

  if (state->g1 == NULL)
    {
      gsl_vector_free (state->x1);
      GSL_ERROR ("failed to allocate space for g1", GSL_ENOMEM);
    }

  state->lr = gsl_vector_alloc (n);

  if (state->lr == NULL)
    {
      gsl_vector_free (state->x1);
      gsl_vector_free (state->g1);
      GSL_ERROR ("failed to allocate space for lr", GSL_ENOMEM);
    }

  return GSL_SUCCESS;
}

static int delta_delta_set (void *vstate, gsl_multimin_function_fdf * fdf,
                      const gsl_vector * x, double *f,
                      gsl_vector * gradient, double step_size, double tol)
{
  delta_delta_state_t *state = (delta_delta_state_t *) vstate;

  GSL_MULTIMIN_FN_EVAL_F_DF (fdf, x, f, gradient);

  state->step = step_size;
  state->default_step = step_size;
  state->tol = tol;
  gsl_vector_set_all(state->lr, 1.0);

  return GSL_SUCCESS;
}


static void delta_delta_free (void *vstate)
{
  delta_delta_state_t *state = (delta_delta_state_t *) vstate;

  gsl_vector_free (state->x1);
  gsl_vector_free (state->g1);
  gsl_vector_free (state->lr);
}

static int delta_delta_restart (void *vstate)
{
  delta_delta_state_t *state = (delta_delta_state_t *) vstate;

  state->step = state->default_step;
  gsl_vector_set_all(state->lr, 1.0);

  return GSL_SUCCESS;
}

#if 0
static int delta_delta_iterate (void *vstate, gsl_multimin_function_fdf * fdf,
                          gsl_vector * x, double *f, gsl_vector * gradient, gsl_vector * dx)
{
    delta_delta_state_t *state = (delta_delta_state_t *) vstate;

    gsl_vector *x1 = state->x1;
    gsl_vector *g1 = state->g1;
    gsl_vector *lr = state->lr;

    size_t i;
    double f0 = *f;
    double f1;
    double step = state->step;
    static const double stepdec = 0.5;
    static const double stepinc = 1.2;
    static const double lrmax = 10.0;
    static const double lrmin = 0.1;
    static const double lrinc = 0.08;
    static const double lrdec = 0.9;

    const size_t lr_size = lr->size;
    const size_t lr_stride = lr->stride;
    const size_t gr_stride = gradient->stride;
    const size_t g1_stride = g1->stride;

  /* compute new trial point at x1= x - step * dir, where dir is the normalized gradient */
    double gnorm = cblas_dnrm2((int)gradient->size, gradient->data, (int)gradient->stride);

    if (gnorm == 0.0) {
        gsl_vector_set_zero (dx);
        return GSL_ENOPROG;
    }

    //gradient : the direction to go in 
    //dx: the step taken
    //x1: x + dx
    //g1: the derivative at the new point x1
    step /= stepdec;
    do {
        step *= stepdec;
        cblas_dcopy((int)dx->size, gradient->data, (int)gradient->stride, dx->data, (int)dx->stride);
        cblas_dscal((int)dx->size, -step / gnorm, dx->data, (int)dx->stride);
        gsl_vector_mul(dx, lr);

        cblas_dcopy((int)x->size, x->data, (int)x->stride, x1->data, (int)x1->stride);
        cblas_daxpy((int)x->size, 1.0, dx->data, (int)dx->stride, x1->data, (int)x1->stride);

        /* evaluate function and gradient at new point x1 */
        GSL_MULTIMIN_FN_EVAL_F_DF (fdf, x1, &f1, g1);

        /* update lr by comparing old gradient with new g1*/
        for (i = 0; i < lr_size; ++i) {
            if ( gradient->data[gr_stride * i] * g1->data[g1_stride * i] > 0 ) {
                lr->data[lr_stride*i] += ( lr->data[lr_stride*i] < lrmax ) ? lrinc : 0.0;
            } else {
                lr->data[lr_stride*i] *= ( lr->data[lr_stride*i] > lrmin ) ? lrdec : 1.0;
            }
        }

    } while (f1 > f0);

    state->step = step * stepinc;
    gsl_vector_memcpy (x, x1);
    gsl_vector_memcpy (gradient, g1);
    *f = f1;

    return GSL_SUCCESS;
}
#else
#include <gsl/gsl_blas.h>
static int delta_delta_iterate (void *vstate, gsl_multimin_function_fdf * fdf,
                          gsl_vector * x, double *f, gsl_vector * gradient, gsl_vector * dx)
{
    delta_delta_state_t *state = (delta_delta_state_t *) vstate;

    gsl_vector *x1 = state->x1;
    gsl_vector *g1 = state->g1;
    gsl_vector *lr = state->lr;

    size_t i;
    double f0 = *f;
    double f1;
    double step = state->step;
    const double stepdec = 0.5;
    const double stepinc = 1.2;
    const double lrmax = 10.0;
    const double lrmin = 0.1;
    const double lrinc = 0.08;
    const double lrdec = 0.9;

  /* compute new trial point at x1= x - step * dir, where dir is the normalized gradient */
    double gnorm = gsl_blas_dnrm2 (gradient);

    if (gnorm == 0.0) {
        gsl_vector_set_zero (dx);
        return GSL_ENOPROG;
    }

    //gradient : the direction to go in 
    //dx: the step taken
    //x1: x + dx
    //g1: the derivative at the new point x1
    step /= stepdec;
    do {
        step *= stepdec;
        gsl_vector_set_zero (dx);
        gsl_blas_daxpy (-step / gnorm, gradient, dx);
        gsl_vector_mul(dx, lr);

        gsl_vector_memcpy (x1, x);
        gsl_blas_daxpy (1.0, dx, x1);

        /* evaluate function and gradient at new point x1 */
        GSL_MULTIMIN_FN_EVAL_F_DF (fdf, x1, &f1, g1);

        /* update lr by comparing old gradient with new g1*/
        for (i = 0; i < lr->size; ++i) {
            if ( gradient->data[gradient->stride * i] * g1->data[g1->stride * i] > 0 ) {
                lr->data[lr->stride*i] += ( lr->data[lr->stride*i] < lrmax ) ? lrinc : 0.0;
            } else {
                lr->data[lr->stride*i] *= ( lr->data[lr->stride*i] > lrmin ) ? lrdec : 1.0;
            }
        }

    } while (f1 > f0);

    state->step = step * stepinc;
    gsl_vector_memcpy (x, x1);
    gsl_vector_memcpy (gradient, g1);
    *f = f1;

    return GSL_SUCCESS;
}
#endif

static const gsl_multimin_fdfminimizer_type delta_delta_type =
{ 
    "delta_delta",  
    sizeof (delta_delta_state_t),
    &delta_delta_alloc,
    &delta_delta_set,
    &delta_delta_iterate,
    &delta_delta_restart,
    &delta_delta_free
};

const gsl_multimin_fdfminimizer_type * msl_multimin_fdfminimizer_deltadelta = &delta_delta_type;

