
/*  msl_madlib.c
 *
 *  Copyright (C) 2005 James Bergstra
 *
 *  This file is part of the Montreal Scientific Library.
 *
 *  The Montreal Scientific Library is free software; you can redistribute it
 *  and/or modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of the License,
 *  or (at your option) any later version.
 *
 *  The Montreal Scientific Library is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along with
 *  the Montreal Scientific Library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

/*
 * decoding functions were taken from the minimad.c sample program in libmad.
 * Here is it's copyright notice:
 */
/*
 * libmad - MPEG audio decoder library
 * Copyright (C) 2000-2004 Underbit Technologies, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: minimad.c,v 1.4 2004/01/23 09:41:32 rob Exp $
 */


#include "msl.h"
#include <mad.h>
#include <assert.h>
#include <sys/stat.h>
#include <sys/mman.h>

//if signal is alloc'ed, it is filled, starting at @offset samples into the song.
//if signal is void, it is alloc'ed to be the size of the song

#define PCM_INITIAL_ALLOC (1<<20)
#define PRINT_MAD_ERROR 0

struct mp3decode 
{
    unsigned char const *mp3;
    unsigned long mp3len;

    char pcm_resize;

    gsl_block * all;
    unsigned long pos;

    unsigned long samplerate;
    unsigned long numchannels;
};
static enum mad_flow input(void *data, struct mad_stream *stream)
{
    struct mp3decode *buffer = (struct mp3decode *)data;

    if ( ! buffer->mp3len ) return MAD_FLOW_STOP;

    mad_stream_buffer(stream, buffer->mp3, buffer->mp3len);

    buffer->mp3len = 0;

    return MAD_FLOW_CONTINUE;
}
static enum mad_flow output(void *data, struct mad_header const *header, struct mad_pcm *pcm)
{
    struct mp3decode *buf = (struct mp3decode *)data;

    const unsigned int nsamples  = pcm->length;
    const unsigned int nchannels = pcm->channels;
    unsigned int i,j;

    if (buf->samplerate == 0) 
    {
        buf->samplerate = pcm->samplerate;
        buf->numchannels = nchannels;
    }
    else 
    {
        if (buf->samplerate != pcm->samplerate)
        {
            if (buf->pos > 44100)
            {
                fprintf(stderr, "Samplerate changing after %lu samples.\n", buf->pos);
                exit(1);
            }
            buf->samplerate = pcm->samplerate;
        }
    }

    //left_ch   = pcm->samples[0];
    //right_ch  = pcm->samples[1];

    for ( i = 0; i < nsamples; ++i)
    {
        //TODO: do this outside the loop!
        if (buf->pos > buf->all->size - nchannels) 
        {
            if (buf->pcm_resize)
            {
                msl_block_realloc( buf->all, buf->all->size * 2);
            } else {
                return MAD_FLOW_STOP;
            }
        }
        for ( j = 0; j < nchannels; ++j)
        {
            buf->all->data[buf->pos] = mad_f_todouble( pcm->samples[j][i] );
            ++buf->pos;
        }
    }
    return MAD_FLOW_CONTINUE;
}
static enum mad_flow error(void *data, struct mad_stream *stream, struct mad_frame *frame)
{

#if PRINT_MAD_ERROR
    struct mp3decode *buffer = (struct mp3decode *)data;
    fprintf(stderr, "decoding error 0x%04x (%s) at byte offset %u\n",
        stream->error, mad_stream_errorstr(stream),
        stream->this_frame - buffer->mp3);
#endif

  /* return MAD_FLOW_BREAK here to stop decoding (and propagate an error) */

  return MAD_FLOW_CONTINUE;
}

int
msl_dsp_block_load_mp3( gsl_block *pcm, ulong * samplerate, size_t * nchannels, const char * path, const int resize )
{
    struct mp3decode buf;
    struct stat istat;

    assert(pcm);
    FILE * f = fopen(path, "r");
    if (NULL == f || fstat(fileno(f), &istat) == -1 || istat.st_size == 0) return -1;

    buf.mp3 = (const unsigned char*)mmap(0, 
            (size_t)istat.st_size,
            PROT_READ, 
            MAP_SHARED, 
            (int)fileno(f),
            (size_t)0);
    if (buf.mp3 == MAP_FAILED) return -2;
    buf.mp3len = istat.st_size;

    buf.pcm_resize = resize;
    buf.all = pcm;
    buf.pos = 0;
    buf.samplerate = 0;

    struct mad_decoder decoder;

    mad_decoder_init(&decoder, &buf, input, 0 /* header */, 0 /* filter */, output, error, 0 /* message */);
    mad_decoder_run(&decoder, MAD_DECODER_MODE_SYNC);
    mad_decoder_finish(&decoder);

    if (munmap((void*)buf.mp3, (size_t)istat.st_size) == -1) return -3;
    fclose(f);

    pcm->size = buf.pos;
    *samplerate = buf.samplerate;
    *nchannels = buf.numchannels;

    return 0; 
}

#if 0
int io_readmp3(const char * path, double ** pcm, unsigned long *samples, unsigned long *samplerate)
{
  struct stat stat;

  FILE * f = fopen(path, "r");
  if (NULL == f || fstat(fileno(f), &stat) == -1 || stat.st_size == 0) return -1;

  struct mp3decode buf;
  buf.mp3 = (const unsigned char*)mmap(0, stat.st_size, PROT_READ, MAP_SHARED, fileno(f), 0);
  if (buf.mp3 == MAP_FAILED) return -2;

  buf.mp3len = stat.st_size;
  buf.pcmalloc = 1024 * 1024;
  buf.pcmlen = 0;
  buf.pcm = (double*)malloc(buf.pcmalloc * sizeof(double));
  buf.samplerate = 0;

  struct mad_decoder decoder;

  mad_decoder_init(&decoder, &buf, input, 0 /* header */, 0 /* filter */, output, error, 0 /* message */);
  mad_decoder_run(&decoder, MAD_DECODER_MODE_SYNC);
  mad_decoder_finish(&decoder);

  if (munmap((void*)buf.mp3, stat.st_size) == -1) return -3;
  fclose(f);

  *pcm = buf.pcm;
  *samples = buf.pcmlen;
  *samplerate = buf.samplerate;
  return 0;
}
#endif
