
/*  msl_nnet_arch.c
 *
 *  Copyright (C) 2005 James Bergstra
 *
 *  This file is part of the Montreal Scientific Library.
 *
 *  The Montreal Scientific Library is free software; you can redistribute it
 *  and/or modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of the License,
 *  or (at your option) any later version.
 *
 *  The Montreal Scientific Library is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along with
 *  the Montreal Scientific Library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include <assert.h>
#include "msl.h"
#define DECL_MALLOC_PTR(t,v) t * v = (t*)malloc(sizeof(t))
#define DECL_CAST(t,v,p) t * v = (t *)( p );

////////////
//
// arch: layer3 softmax

typedef struct t_archlayer3
{
    size_t nhid;
} archlayer3;
typedef struct t_archlayer3buf
{
    gsl_block * buf;

    gsl_matrix_view S;
    gsl_matrix_view S1;
    gsl_matrix_view T;

    gsl_matrix_view S1v, dS1v;
    gsl_matrix_view H, dH;
    gsl_matrix_view H1, dH1;
    gsl_matrix_view H1w, dH1w;
    gsl_matrix_view Y, dY;

    gsl_matrix_view v,dv;
    gsl_matrix_view w,dw;

    gsl_vector_ulong * tmax, *ymax;

    size_t nin;
    size_t nhid;
    size_t nout;

    gsl_rng * r;
} archlayer3buf;

static inline double
_a3l_f( const gsl_vector * x, void * params )
{
    DECL_CAST(archlayer3buf, a3lbuf, params);
    assert(x->stride == 1);
    double * d = x->data;
    d = msl_matrix_view_segment( &a3lbuf->v, d, a3lbuf->nin + 1, a3lbuf->nhid);
    d = msl_matrix_view_segment( &a3lbuf->w, d, a3lbuf->nhid + 1, a3lbuf->nout);

    msl_matrix_gemm_NN(&a3lbuf->S1v.matrix, 0, 1, &a3lbuf->S1.matrix, &a3lbuf->v.matrix);
    msl_nnet_matrix_logistic( &a3lbuf->H.matrix, &a3lbuf->S1v.matrix);
    msl_matrix_gemm_NN(&a3lbuf->H1w.matrix, 0, 1, &a3lbuf->H1.matrix, &a3lbuf->w.matrix);

    msl_nnet_matrix_softmax(&a3lbuf->Y.matrix, &a3lbuf->H1w.matrix);
    return msl_nnet_matrix_cross_entropy(&a3lbuf->T.matrix, &a3lbuf->Y.matrix, NULL, &a3lbuf->dH1w.matrix)
        / a3lbuf->T.matrix.size1;
}
static inline void
_a3l_df( const gsl_vector * x, void * params, gsl_vector * dx )
{
    DECL_CAST(archlayer3buf,a3lbuf, params);
    assert(dx->stride == 1);
    double * d = dx->data;
    d = msl_matrix_view_segment( &a3lbuf->dv, d, a3lbuf->nin + 1, a3lbuf->nhid);
    d = msl_matrix_view_segment( &a3lbuf->dw, d, a3lbuf->nhid + 1, a3lbuf->nout);

    //H1w = H1 x w
    msl_matrix_gemm_NT( &a3lbuf->dH1.matrix, 0, 1, &a3lbuf->dH1w.matrix, &a3lbuf->w.matrix);
    msl_matrix_gemm_TN( &a3lbuf->dw.matrix, 0, 1, &a3lbuf->H1.matrix, &a3lbuf->dH1w.matrix);
    msl_nnet_matrix_logistic_set_deriv(&a3lbuf->dS1v.matrix, &a3lbuf->dH.matrix, &a3lbuf->H.matrix);
    msl_matrix_gemm_TN( &a3lbuf->dv.matrix, 0, 1, &a3lbuf->S1.matrix, &a3lbuf->dS1v.matrix);
}
static void
_a3l_fdf( const gsl_vector * x, void * params, double * f, gsl_vector * df )
{
    *f = _a3l_f(x,params);
    _a3l_df(x, params, df);
}

static inline double
_a3l_f_alt( const gsl_vector * x, void * params )
{
    DECL_CAST(archlayer3buf, a3lbuf, params);
    assert(x->stride == 1);
    double * d = x->data;
    d = msl_matrix_view_segment( &a3lbuf->v, d, a3lbuf->nin + 1, a3lbuf->nhid);
    d = msl_matrix_view_segment( &a3lbuf->w, d, a3lbuf->nhid + 1, a3lbuf->nout);

    msl_matrix_gemm_NN(&a3lbuf->S1v.matrix, 0, 1, &a3lbuf->S1.matrix, &a3lbuf->v.matrix);
    msl_nnet_matrix_logistic( &a3lbuf->H.matrix, &a3lbuf->S1v.matrix);
    msl_matrix_gemm_NN(&a3lbuf->H1w.matrix, 0, 1, &a3lbuf->H1.matrix, &a3lbuf->w.matrix);

//    unnecessary....
//    msl_nnet_matrix_softmax(&a3lbuf->Y.matrix, &a3lbuf->H1w.matrix);

    //msl_matrix_rowmax_index(a3lbuf->l, &a3lbuf->T.matrix); //computed when T
    //is imported
    msl_matrix_rowmax_index(a3lbuf->ymax, &a3lbuf->H1w.matrix);
    return 1.0 - (double)msl_matrix_rowmax_match(&a3lbuf->T.matrix, a3lbuf->tmax, &a3lbuf->H1w.matrix, a3lbuf->ymax)
        / (double) a3lbuf->T.matrix.size1;
}


static archlayer3buf *
a3lbuf_alloc(const size_t nin, const size_t nout, const size_t neg, const size_t nhid)
{
    DECL_MALLOC_PTR(archlayer3buf, a3lbuf);

    size_t bufsize
        = neg * (nin + 1) //S1
        + neg * (nhid * 2) //S1v, dS1v
        + neg * (nhid * 2 + 2 ) //H1, dH1
        + neg * (nout * 2) //H1w, dH1w
        + neg * (nout * 2) //Y
        + neg * (nout); //T

    a3lbuf->buf = gsl_block_alloc(bufsize);
    double * d = a3lbuf->buf->data;
    d = msl_matrix_view_segment( &a3lbuf->S1   , d, neg, nin + 1);
    d = msl_matrix_view_segment( &a3lbuf->S1v  , d, neg, nhid);
    d = msl_matrix_view_segment( &a3lbuf->dS1v , d, neg, nhid);
    d = msl_matrix_view_segment( &a3lbuf->H1   , d, neg, nhid + 1);
    d = msl_matrix_view_segment( &a3lbuf->dH1  , d, neg, nhid + 1);
    d = msl_matrix_view_segment( &a3lbuf->H1w  , d, neg, nout);
    d = msl_matrix_view_segment( &a3lbuf->dH1w , d, neg, nout);
    d = msl_matrix_view_segment( &a3lbuf->Y    , d, neg, nout);
    d = msl_matrix_view_segment( &a3lbuf->dY   , d, neg, nout);
    d = msl_matrix_view_segment( &a3lbuf->T    , d, neg, nout);

    gsl_matrix_set_all( &a3lbuf->S1.matrix, 1.0 );
    gsl_matrix_set_all( &a3lbuf->H1.matrix, 1.0);

    a3lbuf->S = gsl_matrix_submatrix( &a3lbuf->S1.matrix, 0, 0, neg, nin);
    a3lbuf->H = gsl_matrix_submatrix( &a3lbuf->H1.matrix, 0, 0, neg, nhid);
    a3lbuf->dH = gsl_matrix_submatrix( &a3lbuf->dH1.matrix, 0, 0, neg, nhid);

    a3lbuf->tmax = gsl_vector_ulong_alloc(neg);
    a3lbuf->ymax = gsl_vector_ulong_alloc(neg);

    a3lbuf->nin = nin;
    a3lbuf->nhid = nhid;
    a3lbuf->nout = nout;
    a3lbuf->r = gsl_rng_alloc( gsl_rng_default);
    return a3lbuf;
}
static void
a3lbuf_free( archlayer3buf * a3lbuf )
{
    gsl_vector_ulong_free(a3lbuf->tmax);
    gsl_vector_ulong_free(a3lbuf->ymax);

    gsl_block_free( a3lbuf->buf);
    gsl_rng_free( a3lbuf->r);
    free(a3lbuf);
}

static gsl_multimin_function_fdf *
archlayer3_alloc_min(void * aparams, const size_t nin, const size_t nout, const size_t neg)
{
    DECL_CAST( archlayer3, a3l, aparams);
    const size_t nhid = a3l->nhid;

    DECL_MALLOC_PTR(gsl_multimin_function_fdf, fdf);
    fdf->f = _a3l_f;
    fdf->df = _a3l_df;
    fdf->fdf = _a3l_fdf;
    fdf->n = (nin + 1) * nhid + (nhid + 1) * nout;
    fdf->params = a3lbuf_alloc(nin, nout, neg, nhid);
    return fdf;
}
static void
archlayer3_free_min(gsl_multimin_function_fdf * fdf)
{
    a3lbuf_free( (archlayer3buf*) fdf->params);
    free(fdf);
}

static gsl_multimin_function *
archlayer3_alloc_alt(void * aparams, const size_t nin, const size_t nout, const size_t neg)
{
    DECL_CAST( archlayer3, a3l, aparams);
    const size_t nhid = a3l->nhid;

    DECL_MALLOC_PTR(gsl_multimin_function, f);
    f->f = _a3l_f_alt;
    f->n = (nin + 1) * nhid + (nhid + 1) * nout;
    f->params = a3lbuf_alloc(nin, nout, neg, nhid);
    return f;
}
static void
archlayer3_free_alt(gsl_multimin_function * f)
{
    a3lbuf_free( (archlayer3buf*) f->params);
    free(f);
}

static int 
archlayer3_minimize(gsl_multimin_function_fdf * fdf, gsl_multimin_function *f, gsl_vector * wts)
{
    DECL_CAST( archlayer3buf, a3lbuf, fdf->params);
    msl_vector_set_flat(wts, a3lbuf->r, -0.05, 0.05);
    return msl_multimin_best_g_patience(wts, fdf, f, msl_multimin_fdfminimizer_deltadelta, 0.1, 100);
}

static gsl_matrix_view
archlayer3_view_s(void * p)
{
    DECL_CAST( archlayer3buf, a3lbuf, p);
    return a3lbuf->S;
}
static gsl_matrix_view
archlayer3_view_y(void * p)
{
    DECL_CAST( archlayer3buf, a3lbuf, p);
    return a3lbuf->Y;
}
static gsl_matrix_view
archlayer3_view_t(void * p)
{
    DECL_CAST( archlayer3buf, a3lbuf, p);
    return a3lbuf->T;
}
static void
archlayer3_reset(void * p)
{
    DECL_CAST( archlayer3buf, a3lbuf, p);
    msl_matrix_rowmax_index(a3lbuf->tmax, &a3lbuf->T.matrix);
}

msl_nnet_arch * 
msl_nnet_arch_alloc_layer3_softmax(const size_t nhid)
{
    DECL_MALLOC_PTR( archlayer3, p );
    p->nhid = nhid;

    msl_nnet_arch aa = { 
        archlayer3_alloc_min,
        archlayer3_free_min,
        archlayer3_alloc_alt,
        archlayer3_free_alt,
        archlayer3_minimize,
        archlayer3_view_s,
        archlayer3_view_y,
        archlayer3_view_t,
        archlayer3_reset,
        p};
    DECL_MALLOC_PTR( msl_nnet_arch, a);
    *a = aa;
    return a;
}
void
msl_nnet_arch_free_layer3_softmax(msl_nnet_arch * a)
{
    free(a->aparams);
    free(a);
}

